import os
import linuxcnc
import hal, hal_glib
from qtvcp.lib import mdi_text as mdiText
from qtvcp.lib.gcodes import GCodes
from PyQt5 import QtCore, QtWidgets, QtGui
from qtvcp.widgets.mdi_line import MDILine as MDI_WIDGET
from qtvcp.widgets.gcode_editor import GcodeEditor as GCODE
from qtvcp.widgets.tool_offsetview import ToolOffsetView as TOOL_TABLE
from qtvcp.widgets.origin_offsetview import OriginOffsetView as OFFSET_VIEW
from qtvcp.lib.keybindings import Keylookup
from qtvcp.core import Status, Action, Info
from qtvcp import logger
from qtvcp.widgets.stylesheeteditor import  StyleSheetEditor as SSE
from qtvcp.widgets.status_label import StatusLabel as StatLabel

#PHILLS IMPORTS
from PyQt5.QtGui import QFont, QColor
from PyQt5.Qsci import QsciScintilla, QsciLexerCustom, QsciLexerPython
from qtvcp.widgets.gcode_editor import GcodeDisplay as DISPLAY
from qtvcp.widgets.gcode_editor import EditorBase as EDITOR
from qtvcp.widgets.gcode_editor import GcodeLexer as LEXER

LOG = logger.getLogger(__name__)
KEYBIND = Keylookup()
STATUS = Status()
INFO = Info()
ACTION = Action()

class VLine(QtWidgets.QFrame):
    def __init__(self):
        super(VLine, self).__init__()
        self.setFrameShape(self.VLine|self.Plain)

class HandlerClass:
    def __init__(self, halcomp, widgets, paths):
        self.h = halcomp
        self.w = widgets
        self.PATHS = paths
        INIPATH = os.environ.get('INI_FILE_NAME', '/dev/null')
        self.inifile = linuxcnc.ini(INIPATH)
        self.STYLEEDITOR = SSE(widgets,paths)
        self.GCODES = GCodes(widgets)
        self.valid = QtGui.QDoubleValidator(0.0, 999.999, 3)
        


        
        
        KEYBIND.add_call('Key_F12','on_keycall_F12')
        KEYBIND.add_call('Key_F9','on_keycall_F9')
        KEYBIND.add_call('Key_Pause', 'on_keycall_pause')
        KEYBIND.add_call('Key_Plus', 'on_keycall_plus')
        KEYBIND.add_call('Key_Minus', 'on_keycall_minus')
                
        STATUS.connect('general', self.dialog_return)
        STATUS.connect('state-on', lambda w: self.enable_onoff(True))
        STATUS.connect('state-off', lambda w: self.enable_onoff(False))

        STATUS.connect('state-estop', lambda w: self.estop_state(True))
        STATUS.connect('state-estop-reset', lambda w: self.estop_state(False))

#        STATUS.connect('gcode-line-selected', lambda w, line: self.set_start_line(line))
        STATUS.connect('hard-limits-tripped', self.hard_limit_tripped)
#        STATUS.connect('interp-idle', lambda w: self.set_start_line(0))
        STATUS.connect('user-system-changed', self.user_system_changed)
        STATUS.connect('file-loaded', self.file_loaded)
        # STATUS.connect('homed', self.homed)

#        STATUS.connect('all-homed', self.all_homed)
#        STATUS.connect('not-all-homed', lambda w, list: self.set_dro_homed(False))
        STATUS.connect('homed', self.joint_homed)
        STATUS.connect('not-all-homed', self.joint_unhomed)


#        STATUS.connect('graphics-view-changed', self.graphics_view_changed)
        STATUS.connect('g-code-changed', self.gcodes_changed)
        STATUS.connect('m-code-changed', self.mcodes_changed)

        STATUS.connect('program-pause-changed', self.pause_changed) 

        # print('\n******\nprint(dir(self.PATHS)) = {})'.format(dir(self.PATHS)))
        # print('******\nprint(self.CONFIGPATH) = {}'.format(self.PATHS.CONFIGPATH))
        # print('******\nprint(self.PREFS_FILENAME) = {}\n******\n'.format(self.PATHS.PREFS_FILENAME))


# some global variables
        self.axis_list = INFO.AVAILABLE_AXES
#        self.max_linear_velocity = INFO.MAX_LINEAR_VELOCITY
#        self.max_spindle_rpm = INFO.MAX_SPINDLE_SPEED
        self.system_list = ['G53','G54','G55','G56','G57','G58','G59','G59.1','G59.2','G59.3']
        self.slow_jog_factor = 10
#        self.reload_tool = 0
        self.last_loaded_program = 'None'
#        self.first_turnon = True
#        self.onoff_list = ['frame_program', 'dro_frame']
#        self.onoff_list = ['dro_frame', 'jog_frame']
        self.onoff_list = ['dro_frame']
#        self.jog_button_list = ['jog_a_minus', 'jog_a_plus', 'jog_x_minus', 'jog_x_plus', 'jog_y_minus', 'jog_y_plus', 'jog_z_minus', 'jog_z_plus']
#        self.jog_enabled_list = ['a_minus_e.png', 'a_plus_e.png', 'x_minus_e.png', 'x_plus_e.png', 'y_minus_e.png', 'y_plus_e.png', 'z_minus_e.png', 'z_plus_e.png']
#        self.jog_disabled_list = ['a_minus_d.png', 'a_plus_d.png', 'x_minus_d.png', 'x_plus_d.png', 'y_minus_d.png', 'y_plus_d.png', 'z_minus_d.png', 'z_plus_d.png']
        # self.axis_a_list = ['dro_a', 'action_zero_a', 'axistoolbutton_a', 
        #                     'home_a', 'widget_jog_angular', 'widget_increments_angular',
        #                     'a_plus_jogbutton', 'a_minus_jogbutton']
        self.axis_a_list = ['dro_a', 'dro_label_a', 'home_a', 'touch_a', 'jog_a_plus', 'jog_a_minus']
#                            'widget_jog_angular', 'widget_increments_angular']


#        print('AXES:', self.axis_list)

    #     STATUS.connect('homed', lambda w, v: self.summit_homed(w, v))
    #     STATUS.connect('homed', self.summit1_homed)
    # 
    # def summit_homed(self, w, v):
    #     print(w, v)
    # 
    # def summit1_homed(self, w, v):
    #     print('summit1')
    #     print(w, v)

#    def on_resize(self):
#        print('FRAME RESIZED')

    def estop_state(self, state):
        if state:
            print('ESTOP')
        else:
            print('RESET')

    def initialized__(self):
        self.init_pins()
        self.init_preferences()
        self.init_widgets()


        styleFile = '{}/qtplasmac/qtplasmac.qss'.format(self.PATHS.CONFIGPATH)
        with open(styleFile, "r") as fh:
            self.w.setStyleSheet(fh.read())

#        self.w.frame_12.resizeEvent = self.on_resize()
        
#        self.GCODES.setup_list()
        self.w.stackedWidget_log.setCurrentIndex(0)

        self.w.file_reload.clicked.connect(self.file_reload_clicked)
        self.w.jog_slow.clicked.connect(self.jog_slow_clicked)

        self.w.chk_soft_keyboard.stateChanged.connect(self.soft_keyboard)

        self.w.cone_size.valueChanged.connect(self.cone_size_changed)

#        self.w.dro_x.homed.connect(self.dro_homed)


#        self.w.gcodegraphics.percentLoaded.connect(self.percent_loaded)
#        self.w.gcodegraphics.percentDone.connect(self.percent_done)
#        self.w.jog_slow.clicked.connect(self.jog_slow_clicked)

        
        
        kinematics = self.inifile.find('KINS', 'KINEMATICS').lower().replace('=','').replace('trivkins','').replace(' ','') or None
        kinstype = None
        self.coordinates = 'XYZ'
        if 'kinstype' in kinematics:
            kinstype = kinematics.lower().replace(' ','').split('kinstype')[1]
            if 'coordinates' in kinematics:
                kinematics = kinematics.lower().replace(' ','').split('kinstype')[0]
        if 'coordinates' in kinematics:
            self.coordinates = kinematics.split('coordinates')[1].upper()
        # map axes to joints
        for joint in range(len(self.axis_list)):
            if self.axis_list[joint] == 'X':
                self.w.home_x.set_joint(self.coordinates.index(self.axis_list[joint]))
                self.w.home_x.set_joint_number(self.coordinates.index(self.axis_list[joint]))
                self.w.jog_x_minus.set_joint(self.coordinates.index(self.axis_list[joint]))
                self.w.jog_x_plus.set_joint(self.coordinates.index(self.axis_list[joint]))
            elif self.axis_list[joint] == 'Y':
                self.w.home_y.set_joint(self.coordinates.index(self.axis_list[joint]))
                self.w.home_y.set_joint_number(self.coordinates.index(self.axis_list[joint]))
                self.w.jog_y_minus.set_joint(self.coordinates.index(self.axis_list[joint]))
                self.w.jog_y_plus.set_joint(self.coordinates.index(self.axis_list[joint]))
            elif self.axis_list[joint] == 'Z':
                self.w.home_z.set_joint(self.coordinates.index(self.axis_list[joint]))
                self.w.home_z.set_joint_number(self.coordinates.index(self.axis_list[joint]))
                self.w.jog_z_minus.set_joint(self.coordinates.index(self.axis_list[joint]))
                self.w.jog_z_plus.set_joint(self.coordinates.index(self.axis_list[joint]))
            elif self.axis_list[joint] == 'A':
                self.w.home_a.set_joint(self.coordinates.index(self.axis_list[joint]))
                self.w.home_a.set_joint_number(self.coordinates.index(self.axis_list[joint]))
                self.w.jog_a_minus.set_joint(self.coordinates.index(self.axis_list[joint]))
                self.w.jog_a_plus.set_joint(self.coordinates.index(self.axis_list[joint]))
        # hide axis a if not being used
        if 'A' not in self.axis_list:
            for i in self.axis_a_list:
                self.w[i].hide()
#            self.w.lbl_increments_linear.setText('INCREMENTS')


        # see if home all button required
        for joint in range(len(self.coordinates)):
#            print(joint, self.inifile.find('JOINT_{}'.format(joint), 'HOME_SEQUENCE')), type(self.inifile.find('JOINT_{}', 'HOME_SEQUENCE'))
            if not self.inifile.find('JOINT_{}'.format(joint), 'HOME_SEQUENCE'):
#                print(self.inifile.find('JOINT_0', 'HOME_SEQUENCE'))
                self.w.home_all.hide()
                #break

        self.pm_port = self.inifile.find('PLASMAC', 'PM_PORT') or None
        if self.pm_port:
            print('PM_PORT =', self.pm_port)
        else:
            self.w.rs485_frame.hide()
        # self.w.home_stack.setMaximumHeight(len(self.coordinates) * 44 - 4)
        # 
        # if home_all:
        #     self.w.home_stack.setCurrentIndex(1)
        # else:
        #     self.w.home_stack.setCurrentIndex(0)


        STATUS.emit('dro-reference-change-request', 1)


        # print('LAB_X: {}'.format(self.w.dro_label_x.geometry()))
        # print('STACK: {}'.format(self.w.home_stack.geometry()))
        # print(' PAGE: {}'.format(self.w.home_page.geometry()))
        # print('BTN_1: {}'.format(self.w.home_all.geometry()))
        # print('BTN_A: {}'.format(self.w.home_x.geometry()))
#        print('USRBTNS: {}'.format(self.w.frame_user_buttons.geometry()))

#        print('AXIS_LIST: {}    COORDS: {}'.format(self.axis_list, self.coordinates))

    #############################
    # SPECIAL FUNCTIONS SECTION #
    #############################

#    def dro_homed(self):
#        print('DRO HOMED')

    def class_patch__(self):
        self.old_styleText = LEXER.styleText
        LEXER.styleText = self.styleText
        self.old_defaultColor = LEXER.defaultColor
        LEXER.defaultColor = self.defaultColor

    def styleText(self, start, end):
        return

    def defaultColor(self, style):
        return QColor('#fce94f')


    def init_pins(self):
        pass
        # pin = self.h.newpin('spindle_speed_fb', hal.HAL_FLOAT, hal.HAL_IN)
        # hal_glib.GPin(pin).connect('value_changed', self.spindle_fb_changed)
        # pin = self.h.newpin('spindle_amps', hal.HAL_FLOAT, hal.HAL_IN)
        # hal_glib.GPin(pin).connect('value_changed', self.spindle_amps_changed)
        # pin = self.h.newpin('spindle_power', hal.HAL_FLOAT, hal.HAL_IN)
        # hal_glib.GPin(pin).connect('value_changed', self.spindle_power_changed)
        # pin = self.h.newpin('spindle_fault', hal.HAL_U32, hal.HAL_IN)
        # hal_glib.GPin(pin).connect('value_changed', self.spindle_fault_changed)
        # pin = self.h.newpin('modbus-errors', hal.HAL_U32, hal.HAL_IN)
        # hal_glib.GPin(pin).connect('value_changed', self.mb_errors_changed)
        
    def init_preferences(self):
        if not self.w.PREFS_:
            self.add_alarm('CRITICAL - no preference file found, enable preferences in screenoptions widget')
            return
#        self.last_loaded_program = self.w.PREFS_.getpref('last_file_path', None, str,'BOOK_KEEPING')
#        self.reload_tool = self.w.PREFS_.getpref('Tool to load', 0, int,'GUI_OPTIONS')
#        self.w.chk_reload_program.setChecked(self.w.PREFS_.getpref('Reload program', False, bool,'GUI_OPTIONS'))
#        self.w.chk_reload_tool.setChecked(self.w.PREFS_.getpref('Reload tool', False, bool,'GUI_OPTIONS'))
        self.w.chk_keyboard_shortcuts.setChecked(self.w.PREFS_.getpref('Use keyboard shortcuts', False, bool, 'GUI_OPTIONS'))
        self.w.chk_soft_keyboard.setChecked(self.w.PREFS_.getpref('Use soft keyboard', False, bool, 'GUI_OPTIONS'))
#        self.w.chk_run_from_line.setChecked(self.w.PREFS_.getpref('Run from line', False, bool, 'GUI_OPTIONS'))


        rb = self.w.PREFS_.getpref('THC mode', 'thc_auto', str, 'GUI_OPTIONS')
        self.w[rb].setChecked(True)
        self.w.cornerlock_enable.setChecked(self.w.PREFS_.getpref('Corner lock enable', False, bool, 'GUI_OPTIONS'))
        self.w.kerfcross_enable.setChecked(self.w.PREFS_.getpref('Kerf cross enable', False, bool, 'GUI_OPTIONS'))
        self.w.use_auto_volts.setChecked(self.w.PREFS_.getpref('Use auto volts', False, bool, 'GUI_OPTIONS'))
        self.w.ohmic_probe_enable.setChecked(self.w.PREFS_.getpref('Ohmic probe enable', False, bool, 'GUI_OPTIONS'))
        self.w.powermax_enable.setChecked(self.w.PREFS_.getpref('Powermax comms enable', False, bool, 'GUI_OPTIONS'))
        self.w.cone_size.setValue(self.w.PREFS_.getpref('Preview cone size', 0.5, float, 'GUI_OPTIONS'))

        self.soft_keyboard()

        
    def init_widgets(self):
        self.w.main_tab_widget.setCurrentIndex(0)

        self.w.file_open.clicked.connect(lambda : self.w.stack_gcode.setCurrentIndex(1))
#        self.w.file_cancel.clicked.connect(lambda : self.w.stack_gcode.setCurrentIndex(0))
        self.w.file_cancel.clicked.connect(self.file_cancelled)
#        self.w.file_select.clicked.connect(lambda : self.w.stack_gcode.setCurrentIndex(0))

#        STATUS.connect('file_loaded', lambda a, b: self.w.stack_gcode.setCurrentIndex(0))
#        STATUS.connect('file_loaded', self.file_loaded())

#        self.w.actionbutton_rel.setChecked(True)
#        self.w.jog_slider.setMaximum(self.max_linear_velocity * 60)
        self.w.jog_slider.setMaximum(INFO.MAX_LINEAR_VELOCITY * 60)
        self.w.jog_slider.setValue(INFO.DEFAULT_LINEAR_JOG_VEL)
        self.w.slider_jog_angular.setMaximum(INFO.MAX_ANGULAR_JOG_VEL)
        self.w.slider_jog_angular.setValue(INFO.DEFAULT_ANGULAR_JOG_VEL)
#        self.w.slider_feed_ovr.setMaximum(INFO.MAX_FEED_OVERRIDE)
#        self.w.slider_feed_ovr.setValue(100)
#        self.w.slider_rapid_ovr.setMaximum(100)
#        self.w.slider_rapid_ovr.setValue(100)
#        self.w.slider_spindle_ovr.setMinimum(INFO.MIN_SPINDLE_OVERRIDE)
#        self.w.slider_spindle_ovr.setMaximum(INFO.MAX_SPINDLE_OVERRIDE)
#        self.w.slider_spindle_ovr.setValue(100)
        self.w.chk_override_limits.setChecked(False)
        self.w.chk_override_limits.setEnabled(False)
#        self.w.filemanager.show()
#        self.w.file_editor.hide()
#        self.w.lbl_max_rapid.setText(str(self.max_linear_velocity * 60))
        self.w.lbl_search_vel.setText(self.inifile.find('TOOLSENSOR', 'SEARCH_VEL') or '200')
        # self.w.lbl_probe_vel.setText(self.inifile.find('TOOLSENSOR', 'PROBE_VEL') or '50')
        # self.w.lbl_max_probe.setText(self.inifile.find('TOOLSENSOR', 'MAXPROBE') or '10')
        # self.w.lbl_touch_height.setText(self.inifile.find('TOOLSENSOR', 'TOUCH') or '50')
        # self.w.lbl_laser_x.setText(self.inifile.find('LASER', 'X') or '100')
        # self.w.lbl_laser_y.setText(self.inifile.find('LASER', 'Y') or '-10')
        self.w.lbl_home_x.setText(self.inifile.find('JOINT_0', 'HOME') or '50')
        self.w.lbl_home_y.setText(self.inifile.find('JOINT_1', 'HOME') or '50')
        #set up gcode list
#        self.w.gcode_list.currentRowChanged.connect(self.list_row_changed)
#        titles = mdiText.gcode_titles()
#        for key in sorted(titles.iterkeys()):
#            self.w.gcode_list.addItem(key + ' ' + titles[key])
#        self.w.gcode_description.setReadOnly(True)

        self.w.lbl_gcodes = StatLabel()
        self.w.lbl_mcodes = StatLabel()
        self.w.statusbar.addPermanentWidget(VLine())    # <---
        self.w.statusbar.addPermanentWidget(self.w.lbl_gcodes)
        self.w.statusbar.addPermanentWidget(VLine())    # <---
        self.w.statusbar.addPermanentWidget(self.w.lbl_mcodes)


#        print(dir(self.w.gcodegraphics))
#        self.w.gcodegraphics.set_alpha_mode(True)
#        self.w.gcodegraphics.grid_size = 10 / 25.4
        ACTION.SET_GRAPHICS_VIEW('alpha-mode-on')
        # grid size is in inches
        ACTION.SET_GRAPHICS_GRID_SIZE(10 / 25.4)

        self.w.stack_jog.setCurrentIndex(0)
        self.w.stack_gcode.setCurrentIndex(0)
#        self.w.window.setWindowTitle('blah blah blah')
#        for w in self.w.get():
#            print(w)
#        print(dir(self.w))
        self.w.window().setWindowTitle('blah blah blah')
#        print('TITLE:', self.w.windowTitle())

#        print(dir(self.w))
#        print(dir(self.w.gcode_display))
#        print(self.w.gcode_display.children())
#        self.w.gcode_display.setLexer(None)


        # self.w.gcode_display._styles = {
        #     0: 'Default',
        #     1: 'Default',
        #     2: 'Default',
        #     3: 'Default',
        #     4: 'Default',
        #     }
        # for key, value in self.w.gcode_display._styles.iteritems():
        #     setattr(self.w.gcode_display, value, key)
#
        # hack for gcode display as it cannot use qss file
        fore = '#fce94f'
        fore1 = '#ddbb00'
        back = '#1d1d1d'
        back1 = '#2d2d2d'
        back2 = '#6d6d6d'
        font = QFont()
        font.setFamily('Lato')
        font.setFixedPitch(False)
        font.setPointSize(12)
        font.setBold(False)
        self.w.gcode_display.setMarginWidth(0,44)
        self.w.gcode_display.setMarginsFont(font)
        self.w.gcode_display.setMarginsForegroundColor(QColor(fore1))
        self.w.gcode_display.setMarginsBackgroundColor(QColor(back1))
        self.w.gcode_display.setMarkerBackgroundColor(QColor(back2))
#        self.w.gcode_display.setIndicatorForegroundColor(QColor('red'))
#        self.w.gcode_display.setMatchedBraceForegroundColor(QColor(fore))
        self.w.gcode_display.setMatchedBraceBackgroundColor(QColor(back2))
#        self.w.gcode_display.setCaretForegroundColor(QColor('red'))
        self.w.gcode_display.setCaretLineBackgroundColor(QColor(back2))
#        self.w.gcode_display.setSelectionForegroundColor(QColor(fore))
        self.w.gcode_display.setSelectionBackgroundColor(QColor(back2))
        # need both these for the background
        self.w.gcode_display.SendScintilla(QsciScintilla.SCI_STYLESETBACK, QsciScintilla.STYLE_DEFAULT, QColor(back))
        self.w.gcode_display.lexer.setPaperBackground(QColor(back))
        # FIX_ME: find a way to change the font in gcode_editor.py:EditorBase
        #self.w.gcode_display.SendScintilla(QsciScintilla.SCI_STYLESETFONT, 1, 'Lato')
        # self.w.gcode_display.setFont(font)

    def file_cancelled(self):
        self.w.stack_gcode.setCurrentIndex(0)

    def closing_cleanup__(self):
        if not self.w.PREFS_: return
#        self.w.PREFS_.putpref('Tool to load', self.w.statuslabel_tool.text().encode('utf-8'), int, 'GUI_OPTIONS')
#        self.w.PREFS_.putpref('Reload program', self.w.chk_reload_program.isChecked(), bool, 'GUI_OPTIONS')
#        self.w.PREFS_.putpref('Reload tool', self.w.chk_reload_tool.isChecked(), bool, 'GUI_OPTIONS')
        self.w.PREFS_.putpref('Use keyboard shortcuts', self.w.chk_keyboard_shortcuts.isChecked(), bool, 'GUI_OPTIONS')
        self.w.PREFS_.putpref('Use soft keyboard', self.w.chk_soft_keyboard.isChecked(), bool, 'GUI_OPTIONS')
#        self.w.PREFS_.putpref('Run from line', self.w.chk_run_from_line.isChecked(), bool, 'GUI_OPTIONS')

        self.w.PREFS_.putpref('Corner lock enable', self.w.cornerlock_enable.isChecked(), bool, 'GUI_OPTIONS')
        self.w.PREFS_.putpref('Kerf cross enable', self.w.kerfcross_enable.isChecked(), bool, 'GUI_OPTIONS')
        self.w.PREFS_.putpref('Use auto volts', self.w.use_auto_volts.isChecked(), bool, 'GUI_OPTIONS')
        self.w.PREFS_.putpref('Ohmic probe enable', self.w.ohmic_probe_enable.isChecked(), bool, 'GUI_OPTIONS')
        self.w.PREFS_.putpref('Powermax comms enable', self.w.powermax_enable.isChecked(), bool, 'GUI_OPTIONS')
        self.w.PREFS_.putpref('Preview cone size', int(self.w.cone_size.value() *100) / 100, float, 'GUI_OPTIONS')

    def processed_key_event__(self,receiver,event,is_pressed,key,code,shift,cntrl):
        # when typing in MDI, we don't want keybinding to call functions
        # so we catch and process the events directly.
        # We do want ESC, F1 and F2 to call keybinding functions though
        if code not in(QtCore.Qt.Key_Escape,QtCore.Qt.Key_F1 ,QtCore.Qt.Key_F2):
                    #QtCore.Qt.Key_F3,QtCore.Qt.Key_F4,QtCore.Qt.Key_F5):

            # search for the top widget of whatever widget received the event
            # then check if it's one we want the keypress events to go to
            flag = False
            receiver2 = receiver
            while receiver2 is not None and not flag:
                if isinstance(receiver2, QtWidgets.QDialog):
                    flag = True
                    break
                if isinstance(receiver2, QtWidgets.QLineEdit):
                    flag = True
                    break
                if isinstance(receiver2, MDI_WIDGET):
                    flag = True
                    break
                if isinstance(receiver2, GCODE):
                    flag = True
                    break
                if isinstance(receiver2, TOOL_TABLE):
                    flag = True
                    break
                if isinstance(receiver2, OFFSET_VIEW):
                    flag = True
                    break
                receiver2 = receiver2.parent()

            if flag:
                if isinstance(receiver2, GCODE):
                    # if in manual do our keybindings - otherwise
                    # send events to gcode widget
                    if STATUS.is_man_mode() == False:
                        if is_pressed:
                            receiver.keyPressEvent(event)
                            event.accept()
                        return True
                elif is_pressed:
                    receiver.keyPressEvent(event)
                    event.accept()
                    return True
                else:
                    event.accept()
                    return True

        # ok if we got here then try keybindings
        try:
            return KEYBIND.call(self,event,is_pressed,shift,cntrl)
        except NameError as e:
            self.add_alarm('Exception in KEYBINDING: {}'.format (e))
        except Exception as e:
            LOG.error('Exception in KEYBINDING:', exc_info=e)
            print('Error in, or no function for: {} in handler file for-{}'.format(KEYBIND.convert(event),key))
            return False

    #########################
    # CALLBACKS FROM STATUS #
    #########################

    def pause_changed(self, obj, state):
        if state:
#            self.w.action_pause.Text('RESUME')
            self.w.stack_jog.setCurrentIndex(1)
        else:
#            self.w.action_pause.Text('PAUSE')
            self.w.stack_jog.setCurrentIndex(0)

            
    # def spindle_fb_changed(self, data):
    #     print('spindle_fb_changed function')
        # rpm = int(self.h['spindle_speed_fb'])
        # self.w.lbl_spindle_rpm.setText(str(rpm))
        # if float(self.w.label_spindle_set.text()) > self.max_spindle_rpm:
        #     self.w.label_spindle_set.setProperty('overspeed', True)
        # else:
        #     self.w.label_spindle_set.setProperty('overspeed', False)
        # self.w.label_spindle_set.setStyle(self.w.label_spindle_set.style())

    # def spindle_amps_changed(self, data):
    #     print('spindle_amps_changed function')
        # amps = '{:1.1f}'.format(self.h['spindle_amps'])
        # self.w.lbl_spindle_amps.setText(amps)

    # def spindle_power_changed(self, data):
    #     print('spindle_power_changed function')
        # power = '{:4.1f}'.format(self.h['spindle_power'])
        # self.w.lbl_spindle_power.setText(power)

    # def spindle_fault_changed(self, data):
    #     print('spindle_fault_changed function')
        # fault = hex(self.h['spindle_fault'])
        # self.w.lbl_spindle_fault.setText(fault)

    # def mb_errors_changed(self, data):
    #     print('mb_errors_changed function')
        # errors = self.h['modbus-errors']
        # self.w.lbl_mb_errors.setText(str(errors))

    def dialog_return(self, w, message):
        rtn = message.get('RETURN')
        name = message.get('NAME')

        print('\n(dialog_return_function)\nrtn:{}\nname:{}\n'.format(rtn, name))

        # jog_code = bool(message.get('ID') == '_touchoff_')
        # if jog_code and name == 'MESSAGE' and rtn is True:
        #     z_offset = float(self.w.lbl_touch_height.text())
        #     max_probe = self.w.lbl_max_probe.text()
        #     search_vel = self.w.lbl_search_vel.text()
        #     probe_vel = self.w.lbl_probe_vel.text()
        #     ACTION.CALL_MDI('G21 G49')
        #     ACTION.CALL_MDI('G10 L20 P0 Z0')
        #     ACTION.CALL_MDI('G91')
        #     command = 'G38.2 Z-{} F{}'.format(max_probe, search_vel)
        #     if ACTION.CALL_MDI_WAIT(command, 10) == -1: return
        #     if ACTION.CALL_MDI_WAIT('G1 Z4.0') == -1: return
        #     ACTION.CALL_MDI('G4 P0.5')
        #     command = 'G38.2 Z-4.4 F{}'.format(probe_vel)
        #     if ACTION.CALL_MDI_WAIT(command, 10) == -1: return
        #     command = 'G10 L20 P0 Z{}'.format(z_offset)
        #     ACTION.CALL_MDI(command)
        #     command = 'G1 Z10.0 F{}'.format(search_vel)
        #     ACTION.CALL_MDI_WAIT(command, 10)
        #     ACTION.CALL_MDI('G90')

    def user_system_changed(self, obj, data):
        sys = self.system_list[int(data)]
#        self.w.actionbutton_rel.setText(sys)
        self.w.wcs_button.setText('WCS\n{}'.format(sys))

    def file_loaded(self, obj, filename):
        if filename is not None:
            self.w.progressbar.setValue(0)
            self.last_loaded_program = filename
            self.w.stack_gcode.setCurrentIndex(0)
            self.w.label_gcode.setText(os.path.basename(filename))
        else:
            self.add_alarm('Filename not valid')
        print('  LINES:', self.w.gcode_display.lines())

        self.w.window().setWindowTitle('QtPlasmaC')
        print('TITLE:', self.w.windowTitle())

        print('windowFilePath:', self.w.filemanager.windowFilePath())
    # def homed(self, obj, str):
    #     print('homed obj = {} or {}'.format(obj,str))

#        STATUS.connect('homed', self.joint_homed)
#        STATUS.connect('not-all-homed', self.joint_unhomed)

    def joint_homed(self, obj, joint):
        dro = self.coordinates.lower()[int(joint)]
        self.w['home_{}'.format(dro)].set_unhome
        self.w['dro_{}'.format(dro)].setProperty('homed', True)
# TODO - work out how to update the style
#        self.w['dro_{}'.format(dro)].update()
# temp until update sorted
        self.w['dro_{}'.format(dro)].setStyleSheet('QLabel { color: #fce94f}')
        self.w.update

    def joint_unhomed(self, obj, joints):
        for joint in joints:
            dro = self.coordinates.lower()[int(joint)]
            self.w['home_{}'.format(dro)].set_home
            self.w['dro_{}'.format(dro)].setProperty('homed', False)
# TODO - work out how to update the style
#            self.w['dro_{}'.format(dro)].update()
# temp until update sorted
            self.w['dro_{}'.format(dro)].setStyleSheet('QLabel { color: #4d4d4d}')
        if len(joints) < len(self.coordinates):
            self.w.home_all.setEnabled(True)

#    def all_homed(self, obj):

#        self.w.home_all.setEnabled(False)
        # self.set_dro_homed(True)
        # if self.first_turnon is True:
        #     self.first_turnon = False
        #     if self.w.chk_reload_tool.isChecked():
        #         command = 'M61 Q{}'.format(self.reload_tool)
        #         ACTION.CALL_MDI(command)
        #     if self.last_loaded_program and self.w.chk_reload_program.isChecked():
        #         ACTION.OPEN_PROGRAM(self.last_loaded_program)
        #         self.w.filemanager.updateDirectoryView(self.last_loaded_program)

#        self.w.dro_x.setStyleSheet('QLabel { color: #fce94f}')
#        self.w.dro_y.setStyleSheet('QLabel { color: #fce94f}')
#        self.w.dro_z.setStyleSheet('QLabel { color: #fce94f}')
#        self.w.dro_a.setStyleSheet('QLabel { color: #fce94f}')

#         self.w.jog_x_minus.setIcon(QtGui.QIcon('qtplasmac/images/x_minus_e.png'))
#         self.w.jog_x_plus.setIcon(QtGui.QIcon('qtplasmac/images/x_plus_e.png'))
# #        self.w.jog_x_minus.setIconSize(QtCore.QSize(40, 40))
#         self.w.jog_y_minus.setIcon(QtGui.QIcon('qtplasmac/images/y_minus_e.png'))
#         self.w.jog_y_plus.setIcon(QtGui.QIcon('qtplasmac/images/y_plus_e.png'))
#         self.w.jog_z_minus.setIcon(QtGui.QIcon('qtplasmac/images/z_minus_e.png'))
#         self.w.jog_z_plus.setIcon(QtGui.QIcon('qtplasmac/images/z_plus_e.png'))
#         self.w.jog_a_minus.setIcon(QtGui.QIcon('qtplasmac/images/a_minus_e.png'))
#         self.w.jog_a_plus.setIcon(QtGui.QIcon('qtplasmac/images/a_plus_e.png'))

        # self.w.update()

    def hard_limit_tripped(self, obj, tripped, list_of_tripped):
        self.w.chk_override_limits.setEnabled(tripped)
        if not tripped:
            self.w.chk_override_limits.setChecked(False)

#     def graphics_view_changed(self, *str):
# #        print('\nVIEW = {}\n'.format(str))
#         pass

    def gcodes_changed(self, obj, cod):
        self.w.lbl_gcodes.setText('G-Codes: {}'.format(cod))
        
    def mcodes_changed(self, obj, cod):
        self.w.lbl_mcodes.setText('M-Codes: {}'.format(cod))


    #######################
    # CALLBACKS FROM FORM #
    #######################

    def cone_size_changed(self, data):
        self.w.gcodegraphics.set_cone_basesize(data)
        print('CONE:', data)

    def thc_mode_toggled(self, data):
        if data:
            name = self.w.sender().text()
            if self.w.PREFS_:
                self.w.PREFS_.putpref('THC mode', self.w.sender().objectName(), str, 'GUI_OPTIONS')

    def percentLoaded(self, fraction):
        if fraction <1:
            self.w.progressbar.setValue(0)
            self.w.progressbar.setFormat('LOADING COMPLETE')
        else:
            self.w.progressbar.setValue(fraction)
            self.w.progressbar.setFormat('Loading: {}%'.format(fraction))

    def percent_done(self, fraction):
        self.w.progressbar.setValue(fraction)
        if fraction <1:
            self.w.progressbar.setFormat('')
        else:
            self.w.progressbar.setFormat('Completed: {}%'.format(fraction))

#    def home_btn(self, axis):
#        print('homing axis: {}'.format(axis))

    # program frame
#    def btn_start_clicked(self):
#        if self.main_tab_widget.currentIndex() != 0:
#            self.add_alarm('Must be in Preview mode to run a program')
#            return
#        if not STATUS.is_auto_mode():
#            self.add_alarm('Must be in AUTO mode to run a program')
#            return
        # self.w.btn_start.setProperty('running', True)
        # self.w.btn_start.setStyle(self.w.btn_start.style())
#        start_line = int(self.w.lbl_start_line.text().encode('utf-8'))
#        self.add_alarm('Started program from line {}'.format(start_line))
#        ACTION.RUN(start_line)

    def file_reload_clicked(self):
        if self.last_loaded_program != 'None':
            self.w.progressbar.setValue(0)
            ACTION.OPEN_PROGRAM(self.last_loaded_program)

    # tool frame
    # def btn_go_home_clicked(self):
    #     ACTION.CALL_MDI_WAIT('G53 G0 Z0')
    #     command = 'G53 G0 X{} Y{}'.format(self.w.lbl_home_x.text(), self.w.lbl_home_y.text())
    #     ACTION.CALL_MDI_WAIT(command, 10)
 
    # def btn_ref_laser_clicked(self):
    #     print('btn_ref_laser_clicked function')
    #     # command = 'G10 L20 P0 X{} Y{}'.format(self.w.lbl_laser_x.text(), self.w.lbl_laser_y.text())
    #     # ACTION.CALL_MDI(command)

    # def btn_touchoff_clicked(self):
    #     print('btn_touchoff_clicked function')
    #     # if self.w.statuslabel_tool.text() == '0':
    #     #     self.add_alarm('Cannot touchoff with no tool loaded')
    #     #     return
    #     # if not STATUS.is_all_homed():
    #     #     self.add_alarm('Must be homed to perform tool touchoff')
    #     #     return
    #     # max_probe = self.w.lbl_max_probe.text()
    #     # self.add_alarm('Tool touchoff started')
    #     # # instantiate dialog box
    #     # info = 'Jog to within {} mm of touch plate and click OK'.format(max_probe)
    #     # mess = {'NAME':'MESSAGE', 'ID':'_touchoff_', 'MESSAGE':'TOOL TOUCHOFF', 'MORE':info, 'TYPE':'OKCANCEL'}
    #     # STATUS.emit('dialog-request', mess)
        
    # override frame
    def jog_slow_clicked(self, state):
#        slider = self.w.sender().property('slider')
        slider = self.w.jog_slider
        current = slider.value()
        print('SLIDER STUFF', current, type(current), current/self.slow_jog_factor, current*self.slow_jog_factor)
        max = slider.maximum()
        if state:
            self.w.sender().setText('SLOW')
            slider.setMaximum(max / self.slow_jog_factor)
            slider.setValue(current / self.slow_jog_factor)
            slider.setPageStep(10)
        else:
            self.w.sender().setText('FAST')
            slider.setMaximum(max * self.slow_jog_factor)
            slider.setValue(current * self.slow_jog_factor)
            slider.setPageStep(100)

    # file tab
    # def btn_gcode_edit_clicked(self, state):
    #     if not STATUS.is_on_and_idle():
    #         return
    #     for x in ['load', 'next', 'prev']:
    #         self.w['btn_file_{}'.format(x)].setEnabled(not state)
    #     if state:
    #         self.w.filemanager.hide()
    #         self.w.file_editor.show()
    #         self.w.file_editor.editMode()
    #     else:
    #         self.w.filemanager.show()
    #         self.w.file_editor.hide()
    #         self.w.file_editor.readOnlyMode()

    # tool tab
    # def btn_m61_clicked(self):
    #     checked = self.w.tooloffsetview.get_checked_list()
    #     if len(checked) > 1:
    #         self.add_alarm('Select only 1 tool to load')
    #     elif checked:
    #         ACTION.CALL_MDI('M61 Q{}'.format(checked[0]))
    #     else:
    #         self.add_alarm('No tool selected')

    # alarm tab
    # def btn_clear_alarms_clicked(self):
    #     STATUS.emit('update-machine-log', None, 'DELETE')

    # def btn_save_alarms_clicked(self):
    #     text = self.w.machinelog.toPlainText()
    #     filename = self.w.lbl_clock.text().encode('utf-8')
    #     filename = 'alarms_' + filename.replace(' ','_') + '.txt'
    #     with open(filename, 'w') as f:
    #         f.write(text)

    # gcode tab
    # def list_row_changed(self, row):
    #     line = self.w.gcode_list.currentItem().text().encode('utf-8')
    #     text = line.split(' ')[0]
    #     desc = mdiText.gcode_descriptions(text) or 'No Match'
    #     self.w.gcode_description.clear()
    #     self.w.gcode_description.insertPlainText(desc)
    #     if text != 'null':
    #         words = mdiText.gcode_words()
    #         if text in words:
    #             parm = text + ' '
    #             for index, value in enumerate(words[text], start=0):
    #                 parm += value
    #             self.w.gcode_parameters.setText(parm)
    #         else:
    #             self.w.gcode_parameters.setText('')

    # settings tab
    def chk_override_limits_checked(self, state):
        if state:
            print('Override limits set')
            ACTION.SET_LIMITS_OVERRIDE()
        else:
            print('Override limits not set')

#    def chk_run_from_line_checked(self, state):
#        if not state:
#            self.w.lbl_start_line.setText('1')

    #####################
    # GENERAL FUNCTIONS #
    #####################

    def kb_jog(self, state, joint, direction, fast = False, linear = True):
        if not STATUS.is_man_mode() or not STATUS.machine_is_on():
            self.add_alarm('Machine must be ON and in Manual mode to jog')
            return
        if linear:
            distance = STATUS.get_jog_increment()
            rate = STATUS.get_jograte()/60
        else:
            distance = STATUS.get_jog_increment_angular()
            rate = STATUS.get_jograte_angular()/60
        if state:
            if fast:
                rate = rate * 2
            ACTION.JOG(joint, direction, rate, distance)
        else:
            ACTION.JOG(joint, 0, 0, 0)

    # def add_alarm(self, message):
    #     STATUS.emit('update-machine-log', message, 'TIME')

    # def alarm_added(self):
    #     self.w.led_alarm.setState(True)

    # def tab_changed(self, index):
    #     self.w.btn_gcode_edit.setChecked(False)
    #     self.btn_gcode_edit_clicked(False)
    #     if index == 4:
    #         self.w.led_alarm.setState(False)

    # def set_dro_homed(self, state):
    #     print(state)
    #     # print('homed = {}'.format(state))
    #     pass
    #     # self.w.home_all.setProperty('homed', state)
    #     #self.w.home_all.setStyle(self.w.home_all.style())
    #     # for i in map(str.lower, self.axis_list):
    #     #     self.w['dro_{}'.format(i)].setProperty('homed', state)
    #     #     self.w['dro_{}'.format(i)].setStyle(self.w['dro_{}'.format(i)].style())

    def enable_onoff(self, state):
        for widget in self.onoff_list:
            self.w[widget].setEnabled(state)
        # if not state:
        #     for widget in self.jog_button_list:
        #         self.w[widget].setIcon(QtGui.QIcon('qtplasmac/images/{}'.format(self.jog_disabled_list[self.jog_button_list.index(widget)])))
        # if state and STATUS.is_all_homed():
        #     for widget in self.jog_button_list:
        #         self.w[widget].setIcon(QtGui.QIcon('qtplasmac/images/{}'.format(self.jog_enabled_list[self.jog_button_list.index(widget)])))

#    def set_start_line(self, line):
#        if self.w.chk_run_from_line.isChecked():
#            self.w.lbl_start_line.setText(str(line))
#        else:
#            self.w.lbl_start_line.setText('1')
#            self.add_alarm('Run from line is disabled')

    def keyboard_shortcuts(self):
        if self.w.chk_keyboard_shortcuts.isChecked():
            return True
        else:
            self.add_alarm('Keyboard shortcuts are disabled')
            return False

    def soft_keyboard(self):
        if self.w.chk_soft_keyboard.isChecked():
            self.w.mdiline.soft_keyboard_option = True
            input = 'CALCULATOR'
        else:
            self.w.mdiline.soft_keyboard_option = False
            input = 'ENTRY'
        for axis in 'xyza':
            button = 'touch_{}'.format(axis)
            self.w[button].dialog_code = input

    #####################
    # KEY BINDING CALLS #
    #####################

    def on_keycall_ESTOP(self,event,state,shift,cntrl):
        if not event.isAutoRepeat() and state:
            ACTION.SET_ESTOP_STATE(STATUS.estop_is_clear())

    def on_keycall_POWER(self,event,state,shift,cntrl):
        if not event.isAutoRepeat() and state:
            ACTION.SET_MACHINE_STATE(not STATUS.machine_is_on())

    def on_keycall_ABORT(self,event,state,shift,cntrl):
        if not event.isAutoRepeat() and state and STATUS.stat.interp_state != linuxcnc.INTERP_IDLE:
            ACTION.ABORT()

    def on_keycall_HOME(self,event,state,shift,cntrl):
        if state and self.keyboard_shortcuts():
            if STATUS.is_all_homed():
                ACTION.SET_MACHINE_UNHOMED(-1)
            else:
                ACTION.SET_MACHINE_HOMING(-1)

    def on_keycall_pause(self,event,state,shift,cntrl):
        if state and STATUS.is_auto_mode() and self.keyboard_shortcuts():
            ACTION.PAUSE()

    def on_keycall_XPOS(self,event,state,shift,cntrl):
        if not event.isAutoRepeat() and self.keyboard_shortcuts():
            self.kb_jog(state, 0, 1, shift)

    def on_keycall_XNEG(self,event,state,shift,cntrl):
        if not event.isAutoRepeat() and self.keyboard_shortcuts():
            self.kb_jog(state, 0, -1, shift)

    def on_keycall_YPOS(self,event,state,shift,cntrl):
        if not event.isAutoRepeat() and self.keyboard_shortcuts():
            self.kb_jog(state, 1, 1, shift)

    def on_keycall_YNEG(self,event,state,shift,cntrl):
        if not event.isAutoRepeat() and self.keyboard_shortcuts():
            self.kb_jog(state, 1, -1, shift)

    def on_keycall_ZPOS(self,event,state,shift,cntrl):
        if not event.isAutoRepeat() and self.keyboard_shortcuts():
            self.kb_jog(state, 2, 1, shift)

    def on_keycall_ZNEG(self,event,state,shift,cntrl):
        if not event.isAutoRepeat() and self.keyboard_shortcuts():
            self.kb_jog(state, 2, -1, shift)
    
    # def on_keycall_APOS(self,event,state,shift,cntrl):
    #     if not event.isAutoRepeat() and self.keyboard_shortcuts():
    #         self.kb_jog(state, 2, 1, shift)
    # 
    # def on_keycall_ANEG(self,event,state,shift,cntrl):
    #     if not event.isAutoRepeat() and self.keyboard_shortcuts():
    #         self.kb_jog(state, 2, -1, shift)
    
    def on_keycall_plus(self,event,state,shift,cntrl):
        if not event.isAutoRepeat() and self.keyboard_shortcuts():
            self.kb_jog(state, 3, 1, shift, False)

    def on_keycall_minus(self,event,state,shift,cntrl):
        if not event.isAutoRepeat() and self.keyboard_shortcuts():
            self.kb_jog(state, 3, -1, shift, False)

    def on_keycall_F12(self,event,state,shift,cntrl):
        if not event.isAutoRepeat() and state:
            self.STYLEEDITOR.load_dialog()

    def on_keycall_F9(self,event,state,shift,cntrl):
        if not event.isAutoRepeat():
            print('START A MANUAL CUT')

    ##############################
    # required class boiler code #
    ##############################
    def __getitem__(self, item):
        return getattr(self, item)

    def __setitem__(self, item, value):
        return setattr(self, item, value)

################################
# required handler boiler code #
################################

def get_handlers(halcomp, widgets, paths):
    return [HandlerClass(halcomp, widgets, paths)]
